/* -*- mode: Java; c-basic-offset: 4 -*- */
/*------------------------------------------------------------------------
 |  FILE            ExtractRMIInterface.java
 |  WRITTEN BY      Sverre H. Huseby <sverrehu@online.no>
 +----------------------------------------------------------------------*/

import java.io.*;
import java.util.*;
import java.text.*;

/**
 * 
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:sverrehu@online.no">sverrehu@online.no</A>&gt;
 * @version  $Id: ExtractRMIInterface.java,v 1.9 1999/10/19 17:19:46 sverrehu Exp $
 */
public final class ExtractRMIInterface {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private static final String progName       = "ExtractRMIInterface";
    private static final String version        = "1.0.0";

    private static final String magic          = "/* INTERFACE */";
    private static final String magicInterface = "/* INTERFACE-NAME:";
    private static final String magicPackage   = "/* INTERFACE-PACKAGE:";
    private static final String magicImport    = "/* INTERFACE-IMPORT:";
    private static final String magicExtends   = "/* INTERFACE-EXTENDS:";
    private static final String commentEnd     = "*/";

    private boolean verbose;
    private String  outDir;
    private String  interfaceName;
    private String  packageName;
    private Vector  importNames;
    private Vector  extendsNames;
    private String  lineSep;

    private String getTimestamp() {
        SimpleDateFormat isoDateTimeFormat;

        isoDateTimeFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        return isoDateTimeFormat.format((new GregorianCalendar()).getTime());
    }

    private void fatal(String msg) {
	System.err.println(progName + ": " + msg);
	System.exit(1);
    }

    private void verbose(String msg) {
	if (!verbose)
	    return;
	System.out.println(msg);
    }

    /**
     * Finds the index of the first character on the next line.
     */
    private int getIndexPastEndOfLine(String s, int curr) {
        int len;

        len = s.length();
        while (curr < len) {
            if (s.charAt(curr) == '\r') {
                if (curr + 1 < len && s.charAt(curr + 1) == '\n') {
                    /* MS-DOS line terminator */
                    curr += 2;
                    break;
                }
                /* probably Macintosh line terminator */
                ++curr;
                break;
            } else if (s.charAt(curr) == '\n') {
                /* Unix line terminator */
                ++curr;
                break;
            }
            ++curr;
        }
        return curr;
    }

    /**
     * Finds the index of the next character in the string that is
     * not part of a comment.
     */
    private int getNextNonCommentIndex(String s, int curr) {
        int  len, next, idx1, idx2;
        char following;

        ++curr;
        len = s.length();
        if (curr >= len)
            return curr;
        if (s.charAt(curr) != '/')
            return curr;
        next = curr + 1;
        if (next >= len)
            return curr;
        following = s.charAt(next);
        if (following == '/') {
            /* skip to (not past) end of line (include line terminator). */
            idx1 = s.indexOf("\r", curr);
            idx2 = s.indexOf("\n", curr);
            if ((idx2 >= 0 && idx2 < idx1) || idx1 < 0)
                idx1 = idx2;
            curr = idx1;
        } else if (following == '*') {
            /* skip past the first end of comment token. */
            next = s.indexOf("*/", curr);
            if (next < 0)
                curr = len;
            else
                curr = next + 2;
        }
        return curr;
    }

    private int getNonCommentIndexOf(String s, String what, int from) {
        int q, len, whatLen, idx;

        len = s.length();
        whatLen = what.length();
        for (from = getNextNonCommentIndex(s, from - 1);
             from <= len - whatLen;
             from = getNextNonCommentIndex(s, from)) {
            idx = from - 1;
            for (q = 0; q < whatLen; q++) {
                idx = getNextNonCommentIndex(s, idx);
                if (idx != from + q)
                    break;
                if (s.charAt(idx) != what.charAt(q))
                    break;
            }
            if (q == whatLen)
                return from;
        }
        return -1;
    }

    private int getTokenIndex(String s, String token, int from) {
        int  idx, sLen, tokenLen;
        char charBefore, charAfter;

        sLen = s.length();
        tokenLen = token.length();
        while (from < sLen) {
            if ((idx = getNonCommentIndexOf(s, token, from)) < 0)
                return -1;
            if (idx > 0)
                charBefore = s.charAt(idx - 1);
            else
                charBefore = ';';
            if (idx < sLen - tokenLen)
                charAfter = s.charAt(idx + tokenLen);
            else
                charAfter = ';';
            if (!Character.isJavaIdentifierPart(charBefore)
                && !Character.isJavaIdentifierPart(charAfter))
                return idx;
            from = idx + 1;
        }
        return -1;
    }

    /**
     * Finds the first occurrence of a line terminator in the given string,
     * and returns it. If none is found, the system line separator
     * is returned.
     */
    private String getLineTerminator(String s) {
        int  idx1, idx2;
        char c1, c2;

        /* find the first occurrence of any of the two line terminator
         * characters. */
        idx1 = s.indexOf("\r");
        idx2 = s.indexOf("\n");
        if ((idx2 >= 0 && idx2 < idx1) || idx1 < 0)
            idx1 = idx2;

        /* if none is found, return a default. */
        if (idx1 < 0)
            return System.getProperty("line.separator");

        /* now check if the following character is the other one of the
         * two possible line terminators. */
        c1 = s.charAt(idx1);
        if (idx1 == s.length() - 1)
            return s.substring(idx1, idx1 + 1);
        c2 = s.charAt(idx1 + 1);
        if ((c2 == '\r' || c2 == '\n') && c2 != c1)
            return s.substring(idx1, idx1 + 2);
        return s.substring(idx1, idx1 + 1);
    }

    private String vectorToCommaString(Vector v) {
	int    q;
	String ret;

	ret = "";
	synchronized (v) {
	    if (v == null || v.size() == 0)
		return null;
	    for (q = 0; q < v.size(); q++)
		ret += (ret.length() > 0 ? ", " : "")
		    + (String) v.elementAt(q);
	}
	return ret;
    }

    private void makeBackup(String filename) {
	File newFile, bakFile;

	newFile = new File(filename);
	if (!newFile.exists())
	    return;

	/* let's go for the Unix way of doing things. */
	bakFile = new File(filename + "~");
	bakFile.delete(); /* remove old backup */
	if (!newFile.renameTo(bakFile))
	    fatal("unable to create backup file " + bakFile);
    }

    private String replaceOutsideComments(String src, String from, String to) {
	StringBuffer sb;
	int          frm, idx;

	sb = new StringBuffer();
	frm = 0;
	idx = getNonCommentIndexOf(src, from, 0);
	if (idx < 0)
	    return src;

	do {
	    sb.append(src.substring(frm, idx));
	    sb.append(to);
	    frm = idx + from.length();
	    idx = getNonCommentIndexOf(src, from, frm);
	    if (idx < 0)
		idx = src.length();
	} while (frm < src.length());
	
	return sb.toString();
    }

    private Vector vectorSetAdd(Vector to, Vector from) {
	int    q;
	Object o;

	if (from == null)
	    return to;
	synchronized (from) {
	    if (from.size() > 0 && to == null)
		to = new Vector();
	    for (q = 0; q < from.size(); q++) {
		o = from.elementAt(q);
		if (!to.contains(o))
		    to.addElement(o);
	    }
	}
	return to;
    }

    /**
     * Creates a vector containing elements from space- or comma-
     * separated lists from comments staring with the given string.
     */
    private Vector extractVector(String source, String commentStart) {
	int             frm, idx, idxEnd;
	String          s;
	StringTokenizer st;
	Vector          ret;

	ret = new Vector();
	frm = 0;
	while ((idx = source.indexOf(commentStart, frm)) >= 0) {
	    idxEnd = source.indexOf(commentEnd, idx);
	    if (idxEnd < 0)
		fatal("unterminated comment");
	    frm = idxEnd + commentEnd.length();
	    s = source.substring(idx + commentStart.length(), idxEnd);
	    st = new StringTokenizer(s, ", \t\r\n");
	    while (st.hasMoreTokens())
		ret.addElement(st.nextToken());
	}
	if (ret.size() == 0)
	    ret = null;
	return ret;
    }

    private void extractInterface(String source) {
	Vector v;

	v = extractVector(source, magicInterface);
	if (v == null)
	    return;
	if (v.size() > 1)
	    fatal("more than one interface name given");
	if (v.size() == 1 && interfaceName == null)
	    interfaceName = (String) v.elementAt(0);
    }

    private void extractPackage(String source) {
	Vector v;

	v = extractVector(source, magicPackage);
	if (v == null)
	    return;
	if (v.size() > 1)
	    fatal("more than one package name given");
	if (v.size() == 1 && packageName == null)
	    packageName = (String) v.elementAt(0);
    }

    private void extractImport(String source) {
	Vector v;

	v = extractVector(source, magicImport);
	importNames = vectorSetAdd(importNames, v);
    }

    private void extractExtends(String source) {
	Vector v;

	v = extractVector(source, magicExtends);
	extendsNames = vectorSetAdd(extendsNames, v);
    }

    private String makeSureMethodThrowsRemoteException(String s) {
        int idx;

        /* find the throws clause. */
        idx = getTokenIndex(s, "throws", 0);
        
        if (idx < 0) {
            /* no throws clause at all. add one. */
            return s + " throws RemoteException";
        }

        /* have a throws clause. check if RemoteException is
         * present. don't bother doing a very thorough test.
         * if RemoteException is mentioned, that's it. even
         * if it happens to be another than java.rmi.RemoteException.
         * comprende? */
        if (getNonCommentIndexOf(s, "RemoteException", idx) >= 0) {
            /* RemoteException is there. return whatever we got. */
            return s;
        }

        /* no RemoteException. add one. the number 7 is the length
         * of the string "throws ". */
        return s.substring(0, idx) + "throws RemoteException, "
            + s.substring(idx + 7);
    }

    private String getMethods(String inFile, String source) {
	int    idx, idx2, from;
	int    numFound;
	int    sourceLen;
	String whiteSpace, method;
	String ret;

	ret = "";
	numFound = 0;
	sourceLen = source.length();
	idx = 0;
	for (;;) {
	    /* find next magic comment */
	    idx = source.indexOf(magic, idx);
	    if (idx < 0)
		break;

	    /* if the magic comment appears on the same line as the
	     * method definition, we need to prepend the whitespace
	     * that currently is in front of our comment to the method
	     * definition later, so we better find it. */
	    idx2 = idx > 0 ? idx - 1 : idx;
	    while (idx2 >= 0 && Character.isWhitespace(source.charAt(idx2))) {
		if (source.charAt(idx2) == '\r'
		    || source.charAt(idx2) == '\n') {
		    break;
		}
		--idx2;
	    }
	    whiteSpace = source.substring(idx2 + 1, idx);

	    /* skip following whitespace, but not more than up to and
	     * including the next linebreak. */
	    idx += magic.length();
	    while (idx < sourceLen
		   && Character.isWhitespace(source.charAt(idx))) {
		if (source.charAt(idx) == '\r') {
		    if (idx + 1 < sourceLen
			&& source.charAt(idx + 1) == '\n') {
			/* MS-DOS line terminator */
			idx += 2;
			break;
		    }
		    /* probably Macintosh line terminator */
		    ++idx;
		    break;
		} else if (source.charAt(idx) == '\n') {
		    /* Unix line terminator */
		    ++idx;
		    break;
		}
		++idx;
	    }
	    /* idx now points to the first character that should be
	     * included in the output. */
	    if (idx >= sourceLen) {
		System.err.println("WARNING: empty interface element in "
				   + inFile);
		continue;
	    }

	    from = idx;

	    /* include everyting up to but not including the first
	     * occurence of '{' or ';' that is not within a comment */
            char c;
            idx = from - 1;
            for (;;) {
                idx = getNextNonCommentIndex(source, idx);
                if (idx >= sourceLen)
                    break;
                c = source.charAt(idx);
                if (c == '{' || c == ';')
                    break;
            }
            if (idx >= sourceLen)  {
                System.err.println("WARNING: unable to find end of "
                                   + "interface element in " + inFile);
                    continue;
            }

	    /* skip whitespace backwards */
	    --idx;
	    while (idx >= from && Character.isWhitespace(source.charAt(idx)))
		   --idx;
	    if (idx <= from) {
		System.err.println("WARNING: empty interface element in "
				   + inFile);
		continue;
	    }
	    /* clean out any method modifiers that can't be included in
	     * interfaces. we do it rather brutally: method names will
             * be destroyed if they contain these strings. *yikes!* */
	    method = source.substring(from, idx + 1);
	    method = replaceOutsideComments(method, "synchronized ", "");
	    method = replaceOutsideComments(method, "synchronized", "");
	    method = replaceOutsideComments(method, "native ", "");
	    method = replaceOutsideComments(method, "native", "");
	    /* (there are more; static, final, private and protected.) */

            /* all interface methods must throw RemoteException. make
             * sure they do. */
            method = makeSureMethodThrowsRemoteException(method);

	    /* add whatever we found in this round. if what we found
	     * doesn't start with whitespace, assume we should add
	     * the whitespace that was found in front of the magic
	     * comment above. */
	    if (ret.length() > 0)
		ret += lineSep;
	    if (!Character.isWhitespace(source.charAt(from)))
		ret += whiteSpace;
	    ret += method + ";" + lineSep;

	    ++numFound;
	}

	if (numFound == 0)
	    verbose("WARNING: no interface elements found in `"
		    + inFile + "'");
	else
	    verbose(numFound + " interface element"
		    + (numFound != 1 ? "s" : "")
		    + " found in `" + inFile + "'");
	return ret;
    }

    private String getInterface(String methods) {
	int    q;
	String ret;

	ret = "/* this file was generated automatically by "
            + progName + " */";
	ret += lineSep;
        ret += "/* " + getTimestamp() + " */";
	ret += lineSep;
	ret += lineSep;

	if (packageName != null && packageName.length() > 0) {
	    ret += "package " + packageName + ";";
	    ret += lineSep;
	    ret += lineSep;
	}

	for (q = 0; q < importNames.size(); q++) {
	    ret += "import " + (String) importNames.elementAt(q) + ";";
	    ret += lineSep;
	}
	ret += lineSep;

	ret += "public interface " + interfaceName;
	if (extendsNames.size() != 0) {
	    ret += lineSep;
	    ret += "extends " + vectorToCommaString(extendsNames);
	}
	ret += " {";
	ret += lineSep;
	ret += lineSep;
	ret += methods;
	ret += "}";
	ret += lineSep;

	return ret;
    }

    public void save(String checkInFile, String interfaceDefinition)
    throws IOException {
	String     filename;
	FileWriter fw;

	filename = interfaceName + ".java";
	if (outDir != null)
	    filename = (new File(outDir, filename)).toString();
	if (filename.equalsIgnoreCase(checkInFile))
	    fatal("output may overwrite input. check your interface name");
	makeBackup(filename);
	fw = new FileWriter(filename);
	fw.write(interfaceDefinition);
	fw.close();
    }

    private void extract(String inFile) {
	int          c;
	FileReader   fr;
	StringBuffer sb;
	String       source;
	String       methods, interfaceDefinition;
	String       s;

	sb = new StringBuffer();
	try {
	    fr = new FileReader(inFile);
	    while ((c = fr.read()) >= 0)
		sb.append((char) c);
	    source = sb.toString();
	    fr.close();
            lineSep = getLineTerminator(source);
	    extractInterface(source);
	    if (interfaceName == null)
		fatal("no interface name given");
	    extractPackage(source);
	    extractImport(source);
	    extractExtends(source);
	    methods = getMethods(inFile, source);
	    interfaceDefinition = getInterface(methods);
	    verbose("Extracted from `" + inFile + "':");
	    verbose("  interface: " + interfaceName);
	    verbose("  package:   "
		    + (packageName != null ? packageName : "[none]"));
	    s = vectorToCommaString(importNames);
	    verbose("  imports:   " + (s != null ? s : "[none]"));
	    s = vectorToCommaString(extendsNames);
	    verbose("  extends:   " + (s != null ? s : "[none]"));
	    save(inFile, interfaceDefinition);
	} catch (IOException e) {
	    fatal(e.getMessage());
	}
    }

    private String[] removeOption(String[] args, int idx, int len) {
	int      from, to;
	String[] ret;

	if (idx < 0 || idx >= args.length)
	    return args;
	ret = new String[args.length - len];
	for (from = 0, to = 0; from < args.length; from++)
	    if (from < idx || from >= idx + len)
		ret[to++] = args[from];
	return ret;
    }

    private String[] removeOption(String[] args, int idx) {
	return removeOption(args, idx, 1);
    }

    private boolean compareOption(String opt, String shrt, String lng) {
	return ((shrt != null && opt.compareTo("-" + shrt) == 0)
		|| (lng != null && opt.compareTo("-" + lng) == 0));
    }

    private String getStringParameter(String opt, String[] args, int idx) {
	if (idx >= args.length)
	    fatal("missing parameter to option `" + opt + "'");
	return args[idx];
    }

    private String[] parseOptions(String[] args) {
	int    q;
	String opt;

	q = 0;
	while (q < args.length) {
	    opt = args[q];
	    if (!opt.startsWith("-")) {
		++q;
		continue;
	    }
	    if (compareOption(opt, "h", "help")) {
		usage();
		args = removeOption(args, q);
	    } else if (compareOption(opt, "d", "dir")) {
		args = removeOption(args, q);
		outDir = getStringParameter(opt, args, q);
		args = removeOption(args, q);
	    } else if (compareOption(opt, "e", "extends")) {
		args = removeOption(args, q);
		extendsNames.addElement(getStringParameter(opt, args, q));
		args = removeOption(args, q);
	    } else if (compareOption(opt, "i", "import")) {
		args = removeOption(args, q);
		importNames.addElement(getStringParameter(opt, args, q));
		args = removeOption(args, q);
	    } else if (compareOption(opt, "p", "package")) {
		args = removeOption(args, q);
		packageName = getStringParameter(opt, args, q);
		args = removeOption(args, q);
	    } else if (compareOption(opt, "v", "verbose")) {
		args = removeOption(args, q);
		verbose = true;
	    } else
		fatal("unknown option `" + opt + "'");
	}
	return args;
    }

    private void usage() {
	System.out.println(
	  progName + " " + version + " by Sverre H. Huseby\n"
	  + "\n"
	  + "usage: " + progName + " [options] "
	  + "Implementation.java [InterfaceName]\n"
	  + "\n"
	  + "  -d, -dir directory         set output directory\n"
	  + "  -e, -extends interface     add extends for given interface\n"
	  + "  -h, -help                  display this help and exit\n"
	  + "  -i, -import package        include import of given package\n"
	  + "  -p, -package name          set package name\n"
	  + "  -v, -verbose               show messages\n"
	  + "\n"
	);
	System.exit(0);
    }

    private ExtractRMIInterface(String[] args) {
	int q;

	importNames = new Vector();
	importNames.addElement("java.rmi.*");
	extendsNames = new Vector();
	extendsNames.addElement("Remote");
	args = parseOptions(args);

	if (args.length < 1 || args.length > 2)
	    fatal("illegal number of arguments. try -help");

	if (args.length > 1)
	    interfaceName = args[1];

	extract(args[0]);
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/
    public static void main(String[] args) {
	new ExtractRMIInterface(args);
    }
}
